// serial_imu.cpp
#include "serial_parse.h"
#include <thread>
#include <chrono>
#include <fcntl.h>	   // For O_RDWR, O_NOCTTY, O_SYNC
#include <sys/types.h> // For open()
#include <termios.h>   // POSIX terminal control definitions
#include <unistd.h>	   // write(), read(), close()
#include <sys/ioctl.h>
#include <linux/serial.h>

#define IMU_SERIAL "/dev/ttyUSB0"
#define BAUD (115200)
#define BUF_SIZE 4096

uint8_t rx_buffer[BUF_SIZE];

_RINGBUF ringbuf_ch[2];
ros::Publisher IMU_pub;
ros::Publisher IMU_temp_pub;
serial::Serial sp;

uint8_t *ring_pointer(uint8_t chanel)
{
	return ringbuf_ch[chanel].RingBuf;
}

void ring_write(const uint8_t data, uint8_t chanel)
{
	ringbuf_ch[chanel].RingBuf[ringbuf_ch[chanel]._tail++] = data;
	if (ringbuf_ch[chanel]._tail == RingNum)
	{
		ringbuf_ch[chanel]._tail = 0;
	}
}

int ring_available(uint8_t chanel)
{
	int rev_ch = 0;
	rev_ch = ringbuf_ch[chanel]._tail - ringbuf_ch[chanel]._head;
	if (rev_ch < 0)
		rev_ch += RingNum;
	return (rev_ch);
}

uint8_t ring_read(uint8_t chanel)
{
	uint8_t ch;
	ch = ringbuf_ch[chanel].RingBuf[ringbuf_ch[chanel]._head++];
	if (ringbuf_ch[chanel]._head == RingNum)
	{
		ringbuf_ch[chanel]._head = 0;
	}
	return ch;
}
int ring_space(uint8_t chanel)
{
	return RingNum - ring_available(chanel);
}

int configureLowLatencyMode(const std::string &port)
{
	int serial_port = open(port.c_str(), O_RDWR | O_NOCTTY | O_SYNC);
	if (serial_port < 0)
	{
		ROS_ERROR_STREAM("Error opening serial port: " << strerror(errno));
		return -1;
	}

	struct serial_struct ser_info;
	if (ioctl(serial_port, TIOCGSERIAL, &ser_info) < 0)
	{
		ROS_ERROR_STREAM("Error getting serial info: " << strerror(errno));
		close(serial_port);
		return -1;
	}

	ser_info.flags |= ASYNC_LOW_LATENCY;

	if (ioctl(serial_port, TIOCSSERIAL, &ser_info) < 0)
	{
		ROS_ERROR_STREAM("Error setting low latency mode: " << strerror(errno));
		close(serial_port);
		return -1;
	}

	close(serial_port);
	ROS_INFO_STREAM("Low latency mode enabled for port: " << port);
	return 0;
}

void timerCallback(const ros::TimerEvent &)
{
	size_t num = sp.available();
	if (num != 0)
	{
		if (num > BUF_SIZE)
			num = BUF_SIZE;
		num = sp.read(rx_buffer, num);
		for (int i = 0; i < num; i++)
			ring_write(rx_buffer[i], RINGBUF_OF_UART1);
	}
}

int main(int argc, char **argv)
{
	int rev = 0;
	ros::init(argc, argv, "serial_imu");
	ros::NodeHandle n;

	IMU_pub = n.advertise<sensor_msgs::Imu>("/IMU_data", 20);
	IMU_temp_pub= n.advertise<sensor_msgs::Temperature>("/IMU_temp", 20);
	serial::Timeout to = serial::Timeout::simpleTimeout(10);
	// Configure low latency mode
	if (configureLowLatencyMode(IMU_SERIAL) < 0)
	{
		return -1;
	}

	sp.setPort(IMU_SERIAL);

	sp.setBaudrate(BAUD);

	sp.setTimeout(to);

	try
	{
		sp.open();
	}
	catch (serial::IOException &e)
	{
		ROS_ERROR_STREAM("Unable to open port.");
		return -1;
	}

	if (sp.isOpen())
	{
		ROS_INFO_STREAM("/dev/ttyUSB0 is opened.");
	}
	else
	{
		return -1;
	}
	sp.flushInput();  // 清空接收缓冲区
	sp.flushOutput(); // 清空发送缓冲区

	// Send_CMD_LONG(14,101,0,4,0,0,0); //设置IMU坐标系，默认为101，前右下
	// for (int i = 0; i < 10; i++)
	// sp.write((uint8_t *)&data_cmd_long,sizeof(data_cmd_long));

	Send_CMD_LONG(3, 1, 0, 0, 0, 0, 0); // 获取AHRS数据流
	for (int i = 0; i < 10; i++)
		sp.write((uint8_t *)&data_cmd_long, sizeof(data_cmd_long));

	Send_CMD_LONG(5, 0, 0, 0, 0, 0, 0); // 保存到flash
	for (int i = 0; i < 3; i++)
		sp.write((uint8_t *)&data_cmd_long, sizeof(data_cmd_long));

	Send_CMD_LONG(3, 1, 0, 0, 0, 0, 0); // 再次获取AHRS数据流
	for (int i = 0; i < 10; i++)
		sp.write((uint8_t *)&data_cmd_long, sizeof(data_cmd_long));
	sp.flushInput(); // 只清除接收缓存
	ros::Timer timer = n.createTimer(ros::Duration(0.001), timerCallback);
	uint8_t rx_data;
	while (ros::ok())
	{
		ros::spinOnce();
		while (ring_available(RINGBUF_OF_UART1))
		{
			rx_data = ring_read(RINGBUF_OF_UART1);
			imu_rx(rx_data);
		}
		std::this_thread::sleep_for(std::chrono::milliseconds(10));
	}

	sp.close();

	return 0;
}
